﻿/*
 * Extra components rendering
 *
 * Author   : Mikhail Shiryaev
 * Created  : 2018
 * Modified : 2019
 *
 * Requires:
 * - jquery
 * - schemecommon.js
 * - schemerender.js
 */

/********** Extra Components Utilities **********/

scada.scheme.extraCompUtils = {
    // Calculate a proportion of the specified value between minimum and maximum
    calcProportion: function (val, min, max) {
        var proportion = min < max ? (val - min) / (max - min) : 0;

        if (proportion < 0) {
            proportion = 0;
        } else if (proportion > 1) {
            proportion = 1;
        }

        return proportion;
    },

    // Normalize angle to the range from 0 to 2*Pi
    normAngle: function (angle) {
        var doublePi = 2 * Math.PI;

        while (angle < 0) {
            angle += doublePi;
        }

        while (angle >= doublePi) {
            angle -= doublePi;
        }

        return angle;
    }
};

/********** Gauge Renderer **********/

scada.scheme.GaugeRenderer = function () {
    this.GAUGE_BACK_IMAGE = "url('../SchExtraComp/images/gauge-back.svg')";
    this.GAUGE_POINTER_IMAGE = "url('../SchExtraComp/images/gauge-pointer.svg')";

    scada.scheme.StaticPictureRenderer.call(this);
};

scada.scheme.GaugeRenderer.prototype = Object.create(scada.scheme.StaticPictureRenderer.prototype);
scada.scheme.GaugeRenderer.constructor = scada.scheme.GaugeRenderer;

scada.scheme.GaugeRenderer.prototype._getCanvasFont = function (fontProp) {
    // underline not supported
    return (fontProp.Italic ? "italic " : "") + (fontProp.Bold ? "bold " : "") +
        fontProp.Size + "px " + fontProp.Name;
};

scada.scheme.GaugeRenderer.prototype._createScale = function (props) {
    var min = props.Minimum;
    var max = props.Maximum;
    var valRange = max - min;
    var scaleProps = props.Scale;
    var step = scaleProps.Step;
    var showLabels = scaleProps.ShowLabels;
    var showTicks = scaleProps.ShowTicks;

    if (valRange > 0 && step > 0 && (showLabels || showTicks)) {
        var cnvScale = $("<canvas class='extra-gauge-scale'></canvas>");

        cnvScale.css({
            "width": props.Size.Width,
            "height": props.Size.Height
        });

        var canvas = cnvScale[0];
        canvas.width = props.Size.Width;
        canvas.height = props.Size.Height;

        var context = canvas.getContext("2d");
        context.textAlign = "center";
        context.textBaseline = "middle";
        context.fillStyle = scaleProps.ForeColor;

        if (props.Scale.Font) {
            context.font = this._getCanvasFont(scaleProps.Font);
        }

        var minAngle = (90 - props.MinAngle) * Math.PI / 180; // in radians
        var angleRange = (props.MaxAngle - props.MinAngle) * Math.PI / 180;
        var decimalPlaces = scaleProps.DecimalPlaces;
        var lableRadius = scaleProps.LabelRadius;
        var rotateLabels = scaleProps.RotateLabels;
        var tickRadius = scaleProps.TickRadius;
        var tickWidth = scaleProps.TickSize.Width;
        var halfTickWidth = tickWidth / 2;
        var tickHeight = scaleProps.TickSize.Height;
        var pivotX = props.PivotLocation.X;
        var pivotY = props.PivotLocation.Y;
        var halfPi = Math.PI / 2;

        for (var val = min; val <= max; val += step) {
            var angle = minAngle - angleRange * (val - min) / valRange;
            var cos = Math.cos(angle);
            var sin = Math.sin(angle);

            if (showLabels) {
                var text = val.toFixed(decimalPlaces);
                var labelX = Math.round(cos * lableRadius) + pivotX;
                var labelY = Math.round(-sin * lableRadius) + pivotY;

                if (rotateLabels) {
                    var normAngle = scada.scheme.extraCompUtils.normAngle(angle);
                    var labelAngle = halfPi - normAngle;

                    if (normAngle > Math.PI) {
                        labelAngle += Math.PI;
                    }

                    context.translate(labelX, labelY);
                    context.rotate(labelAngle);
                    context.fillText(text, 0, 0);
                    context.rotate(-labelAngle);
                    context.translate(-labelX, -labelY);
                } else {
                    context.fillText(text, labelX, labelY);
                }
            }

            if (showTicks) {
                var tickX = Math.round(cos * tickRadius) + pivotX;
                var tickY = Math.round(-sin * tickRadius) + pivotY;
                var tickAngle = halfPi - angle;

                context.translate(tickX, tickY);
                context.rotate(tickAngle);
                context.fillRect(-halfTickWidth, 0, tickWidth, tickHeight);
                context.rotate(-tickAngle);
                context.translate(-tickX, -tickY);
            }
        }

        return cnvScale;
    } else {
        return $();
    }
};

scada.scheme.GaugeRenderer.prototype._createPointer = function (props, renderContext) {
   var ImageStretches = scada.scheme.ImageStretches;
   var divPointer = $("<div class='extra-gauge-pointer'></div>");

    divPointer.css({
        "width": props.Pointer.Size.Width,
        "height": props.Pointer.Size.Height,
        "top": props.PivotLocation.Y - props.Pointer.CapLocation.Y,
        "left": props.PivotLocation.X - props.Pointer.CapLocation.X,
        "transform-origin": props.Pointer.CapLocation.X + "px " + props.Pointer.CapLocation.Y + "px",
        "background-repeat": "no-repeat",
        "background-position": "center",
        "background-clip": "padding-box"
    });

    switch (props.Pointer.ImageStretch) {
        case ImageStretches.FILL:
            divPointer.css("background-size",
                props.Pointer.Size.Width + "px " + props.Pointer.Size.Height + "px");
            break;
        case ImageStretches.ZOOM:
            divPointer.css("background-size", "contain");
            break;
    }

    if (props.Pointer.ImageName) {
        var pointerImage = renderContext.getImage(props.Pointer.ImageName);
        this.setBackgroundImage(divPointer, pointerImage);
    } else {
        divPointer.css("background-image", this.GAUGE_POINTER_IMAGE);
    }

    return divPointer;
};

scada.scheme.GaugeRenderer.prototype.createDom = function (component, renderContext) {
    scada.scheme.StaticPictureRenderer.prototype.createDom.call(this, component, renderContext);

    var props = component.props;
    var divComp = component.dom;
    this.bindAction(divComp, component, renderContext);
    divComp.addClass("extra-gauge");

    if (!props.ImageName) {
        divComp.css("background-image", this.GAUGE_BACK_IMAGE);
    }

    if (!renderContext.editMode) {
        divComp.addClass("undef");
    }

    var divContainer = $("<div class='extra-gauge-container'></div>");
    var scaleElem = this._createScale(props);
    var pointerElem = this._createPointer(props, renderContext);

    divContainer.append(scaleElem);
    divContainer.append(pointerElem);
    divComp.append(divContainer);
};

scada.scheme.GaugeRenderer.prototype.setSize = function (component, width, height) {
    scada.scheme.ComponentRenderer.prototype.setSize.call(this, component, width, height);

    // replace the scale
    var divComp = component.dom;
    var oldScaleElem = divComp.children(".extra-gauge-scale");
    var newScaleElem = this._createScale(component.props);
    oldScaleElem.replaceWith(newScaleElem);
};

scada.scheme.GaugeRenderer.prototype.updateData = function (component, renderContext) {
    var props = component.props;

    if (props.InCnlNum > 0) {
        var divComp = component.dom;
        var cnlDataExt = renderContext.getCnlDataExt(props.InCnlNum);

        if (cnlDataExt.Stat > 0) {
            // rotate the pointer
            divComp.removeClass("undef");
            var divPointer = divComp.find(".extra-gauge-pointer:first");
            var proportion = scada.scheme.extraCompUtils.calcProportion(cnlDataExt.Val, props.Minimum, props.Maximum);
            var angle = (props.MaxAngle - props.MinAngle) * proportion + props.MinAngle;
            divPointer.css("transform", "rotate(" + angle + "deg)");
        } else {
            divComp.addClass("undef");
        }
    }
};

/********** Level Renderer **********/

scada.scheme.LevelRenderer = function () {
    scada.scheme.ComponentRenderer.call(this);
};

scada.scheme.LevelRenderer.prototype = Object.create(scada.scheme.ComponentRenderer.prototype);
scada.scheme.LevelRenderer.constructor = scada.scheme.LevelRenderer;

scada.scheme.LevelRenderer.prototype._applySize = function (divComp, divContainer, divBar, component) {
    var props = component.props;
    var borders = (props.BorderWidth + props.Padding) * 2;
    var isVertical = props.Size.Width <= props.Size.Height;
    var levelWidth = props.Size.Width - borders;
    var levelHeight = props.Size.Height - borders;

    divComp.css({
        "padding": props.Padding
    });

    divContainer.css({
        "width": levelWidth,
        "height": levelHeight,
        "min-width": levelWidth,
        "min-height": levelHeight
    });

    if (isVertical) {
        divBar.css({
            "width": levelWidth,
            "height": levelHeight / 2
        });
    } else {
        divBar.css({
            "width": levelWidth / 2,
            "height": levelHeight
        });
    }
};

scada.scheme.LevelRenderer.prototype.createDom = function (component, renderContext) {
    var props = component.props;

    var divComp = $("<div id='comp" + component.id + "' class='extra-level'></div>");
    var divContainer = $("<div class='extra-level-container'></div>");
    var divBar = $("<div class='extra-level-bar'></div>");

    this.prepareComponent(divComp, component, true);
    this.bindAction(divComp, component, renderContext);
    this.setBackColor(divBar, props.BarColor);
    this._applySize(divComp, divContainer, divBar, component);

    if (!renderContext.editMode) {
        divComp.addClass("undef");
    }

    divContainer.append(divBar);
    divComp.append(divContainer);
    component.dom = divComp;
};

scada.scheme.LevelRenderer.prototype.setSize = function (component, width, height) {
    scada.scheme.ComponentRenderer.prototype.setSize.call(this, component, width, height);

    var divComp = component.dom;
    var divContainer = divComp.children(".extra-level-container");
    var divBar = divContainer.children(".extra-level-bar");
    this._applySize(divComp, divContainer, divBar, component);
};

scada.scheme.LevelRenderer.prototype.updateData = function (component, renderContext) {
    var props = component.props;
    component.cmdVal = 0;

    if (props.InCnlNum > 0) {
        var divComp = component.dom;
        var divBar = divComp.find(".extra-level-bar");
        var cnlDataExt = renderContext.getCnlDataExt(props.InCnlNum);

        if (cnlDataExt.Stat > 0) {
            divComp.removeClass("undef");

            var proportion = scada.scheme.extraCompUtils.calcProportion(cnlDataExt.Val, props.Minimum, props.Maximum);
            var isVertical = props.Size.Width <= props.Size.Height;
            var borders = (props.BorderWidth + props.Padding) * 2;

            if (isVertical) {
                divBar.css("height", (props.Size.Height - borders) * proportion);
            } else {
                divBar.css("width", (props.Size.Width - borders) * proportion);
            }
        } else {
            divComp.addClass("undef");
        }

        // set colors that depend on status
        var statusColor = cnlDataExt.Color;
        this.setBackColor(divComp, props.BackColor, true, statusColor)
        this.setBorderColor(divComp, props.BorderColor, true, statusColor)
        this.setBackColor(divBar, props.BarColor, true, statusColor)
    }
};

/********** Renderer Map **********/

// Add components to the renderer map
scada.scheme.rendererMap.set("Scada.Web.Plugins.SchExtraComp.Gauge", new scada.scheme.GaugeRenderer());
scada.scheme.rendererMap.set("Scada.Web.Plugins.SchExtraComp.Level", new scada.scheme.LevelRenderer());
